package org.springframework.boot.loader;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.JarURLConnection;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLConnection;
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.util.Enumeration;
import java.util.function.Supplier;
import java.util.jar.JarFile;
import java.util.jar.Manifest;
import org.springframework.boot.loader.archive.Archive;
import org.springframework.boot.loader.jar.Handler;
import org.springframework.boot.loader.jar.JarFile;

public class LaunchedURLClassLoader extends URLClassLoader {
  private static final int BUFFER_SIZE = 4096;
  
  private final boolean exploded;
  
  private final Archive rootArchive;
  
  static {
    ClassLoader.registerAsParallelCapable();
  }
  
  private final Object packageLock = new Object();
  
  private volatile DefinePackageCallType definePackageCallType;
  
  public LaunchedURLClassLoader(URL[] urls, ClassLoader parent) {
    this(false, urls, parent);
  }
  
  public LaunchedURLClassLoader(boolean exploded, URL[] urls, ClassLoader parent) {
    this(exploded, null, urls, parent);
  }
  
  public LaunchedURLClassLoader(boolean exploded, Archive rootArchive, URL[] urls, ClassLoader parent) {
    super(urls, parent);
    this.exploded = exploded;
    this.rootArchive = rootArchive;
  }
  
  public URL findResource(String name) {
    if (this.exploded)
      return super.findResource(name); 
    Handler.setUseFastConnectionExceptions(true);
    try {
      return super.findResource(name);
    } finally {
      Handler.setUseFastConnectionExceptions(false);
    } 
  }
  
  public Enumeration<URL> findResources(String name) throws IOException {
    if (this.exploded)
      return super.findResources(name); 
    Handler.setUseFastConnectionExceptions(true);
    try {
      return (Enumeration<URL>)new UseFastConnectionExceptionsEnumeration(super.findResources(name));
    } finally {
      Handler.setUseFastConnectionExceptions(false);
    } 
  }
  
  protected Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
    if (name.startsWith("org.springframework.boot.loader.jarmode."))
      try {
        Class<?> result = loadClassInLaunchedClassLoader(name);
        if (resolve)
          resolveClass(result); 
        return result;
      } catch (ClassNotFoundException classNotFoundException) {} 
    if (this.exploded)
      return super.loadClass(name, resolve); 
    Handler.setUseFastConnectionExceptions(true);
    try {
      try {
        definePackageIfNecessary(name);
      } catch (IllegalArgumentException ex) {
        if (getPackage(name) == null)
          throw new AssertionError("Package " + name + " has already been defined but it could not be found"); 
      } 
      return super.loadClass(name, resolve);
    } finally {
      Handler.setUseFastConnectionExceptions(false);
    } 
  }
  
  private Class<?> loadClassInLaunchedClassLoader(String name) throws ClassNotFoundException {
    String internalName = name.replace('.', '/') + ".class";
    InputStream inputStream = getParent().getResourceAsStream(internalName);
    if (inputStream == null)
      throw new ClassNotFoundException(name); 
    try {
      try {
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        byte[] buffer = new byte[4096];
        int bytesRead = -1;
        while ((bytesRead = inputStream.read(buffer)) != -1)
          outputStream.write(buffer, 0, bytesRead); 
        inputStream.close();
        byte[] bytes = outputStream.toByteArray();
        Class<?> definedClass = defineClass(name, bytes, 0, bytes.length);
        definePackageIfNecessary(name);
        return definedClass;
      } finally {
        inputStream.close();
      } 
    } catch (IOException ex) {
      throw new ClassNotFoundException("Cannot load resource for class [" + name + "]", ex);
    } 
  }
  
  private void definePackageIfNecessary(String className) {
    int lastDot = className.lastIndexOf('.');
    if (lastDot >= 0) {
      String packageName = className.substring(0, lastDot);
      if (getPackage(packageName) == null)
        try {
          definePackage(className, packageName);
        } catch (IllegalArgumentException ex) {
          if (getPackage(packageName) == null)
            throw new AssertionError("Package " + packageName + " has already been defined but it could not be found"); 
        }  
    } 
  }
  
  private void definePackage(String className, String packageName) {
    try {
      AccessController.doPrivileged(() -> {
            String packageEntryName = packageName.replace('.', '/') + "/";
            String classEntryName = className.replace('.', '/') + ".class";
            for (URL url : getURLs()) {
              try {
                URLConnection connection = url.openConnection();
                if (connection instanceof JarURLConnection) {
                  JarFile jarFile = ((JarURLConnection)connection).getJarFile();
                  if (jarFile.getEntry(classEntryName) != null && jarFile.getEntry(packageEntryName) != null && jarFile.getManifest() != null) {
                    definePackage(packageName, jarFile.getManifest(), url);
                    return null;
                  } 
                } 
              } catch (IOException iOException) {}
            } 
            return null;
          }AccessController.getContext());
    } catch (PrivilegedActionException privilegedActionException) {}
  }
  
  protected Package definePackage(String name, Manifest man, URL url) throws IllegalArgumentException {
    if (!this.exploded)
      return super.definePackage(name, man, url); 
    synchronized (this.packageLock) {
      return doDefinePackage(DefinePackageCallType.MANIFEST, () -> super.definePackage(name, man, url));
    } 
  }
  
  protected Package definePackage(String name, String specTitle, String specVersion, String specVendor, String implTitle, String implVersion, String implVendor, URL sealBase) throws IllegalArgumentException {
    if (!this.exploded)
      return super.definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, sealBase); 
    synchronized (this.packageLock) {
      if (this.definePackageCallType == null) {
        Manifest manifest = getManifest(this.rootArchive);
        if (manifest != null)
          return definePackage(name, manifest, sealBase); 
      } 
      return doDefinePackage(DefinePackageCallType.ATTRIBUTES, () -> super.definePackage(name, specTitle, specVersion, specVendor, implTitle, implVersion, implVendor, sealBase));
    } 
  }
  
  private Manifest getManifest(Archive archive) {
    try {
      return (archive != null) ? archive.getManifest() : null;
    } catch (IOException ex) {
      return null;
    } 
  }
  
  private <T> T doDefinePackage(DefinePackageCallType type, Supplier<T> call) {
    DefinePackageCallType existingType = this.definePackageCallType;
    try {
      this.definePackageCallType = type;
      return call.get();
    } finally {
      this.definePackageCallType = existingType;
    } 
  }
  
  public void clearCache() {
    if (this.exploded)
      return; 
    for (URL url : getURLs()) {
      try {
        URLConnection connection = url.openConnection();
        if (connection instanceof JarURLConnection)
          clearCache(connection); 
      } catch (IOException iOException) {}
    } 
  }
  
  private void clearCache(URLConnection connection) throws IOException {
    Object jarFile = ((JarURLConnection)connection).getJarFile();
    if (jarFile instanceof JarFile)
      ((JarFile)jarFile).clearCache(); 
  }
}
